---
title: Ruby 2.4 changes
prev: 2.5
description: Ruby 2.4 full and annotated changelog
---

# Ruby 2.4

<!--
* **Details:**
* **Reason:**
* **Discussion:** [Feature #](https://bugs.ruby-lang.org/issues/)
* **Documentation:** []()
* **Code:**
* **Follow-up:**
 -->

* **Released at:** Dec 25, 2016 (<a class="github" href="https://github.com/ruby/ruby/blob/trunk/doc/NEWS-2.4.0">NEWS</a> file)
* **Status (as of Dec 25, 2023):** EOL, latest is 2.4.10
* **This document first published:** Oct 14, 2019
* **Last change to this document:** Dec 25, 2023

## Highlights[](#highlights)

* [Toplevel `return`](#toplevel-return)
* [Unification of `Fixnum` and `Bignum` into `Integer`](#fixnum-and-bignum-are-unified-into-integer)
* [Full Unicode support for String case operations](#unicode-case-conversions)
* [`Warning` module for fine-grained warnings output control](#warning-module)
* [`Comparable#clamp`](#comparableclamp)

## Language[](#language)

### Multiple assignment allowed in conditional expression[](#multiple-assignment-allowed-in-conditional-expression)

* **Reason:** It is not what typically considered good style, but multiple assignment in conditions is now consistent with outside of condition behavior.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/10617">Feature #10617</a>
* **Documentation:** —
* **Code:**
  ```ruby
  points = []
  if (x, y = points.first) # in Ruby <2.4, will raise SyntaxError: multiple assignment in conditional
    p [x, y]
  end
  ```
* **Note:** Be aware that condition considered false if the _whole right hand_ of assignment is `false`/`nil`, not the first assigned variable:
  ```ruby
  points = [[]]
  if (x, y = points.first) # x is nil, y is nil, but condition is [], which is truthy
    p [x, y] # will print this
  end
  ```

### Toplevel `return`[](#toplevel-return)

`return` statement at the top level of any `.rb` file stops further execution of this file.

* **Reason:** Useful for code that depends on platform, presence of third-party libraries and so on; `return unless some_condition` at the beginning of the file will allow writing the rest of the code in assumption that necessary condition satisfied.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/4840">Feature #4840</a>
* **Documentation:** —
* **Code:**
  ```ruby
  # Some nokogiri_patch.rb

  # In Ruby < 2.4:
  if defined? Nokogiri
    # the rest of the code should all be nested in the unless
  end

  # Ruby 2.4
  return unless defined? Nokogiri

  # The rest of the code can be written at the top level, now.
  ```

## Refinements improvements[](#refinements-improvements)

_NB: Some of those changes are language-level, some are just method changes, but all a related to refinements._

### Refinements are supported in `Symbol#to_proc` and `send`[](#refinements-are-supported-in-symbolto_proc-and-send)

* **Discussions:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/9451">Feature #9451</a>, <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11476">Feature #11476</a>
* **Code:**
  ```ruby
  module Tests
    refine Numeric do
      def normalize100
        clamp(0, 100)
      end
    end
  end

  using Tests

  p [1, 700, 132].map(&:normalize100)
  # Ruby 2.3: undefined method `normalize100' for 1
  # Ruby 2.4: => [1, 100, 100]

  p 123.send(:normalize100)
  # Ruby 2.3: undefined method `normalize100' for 1
  # Ruby 2.4: => 100
  ```

### `refine` can refine modules, too[](#refine-can-refine-modules-too)

Previously, only classes could've been refined.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12534">Feature #12534</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Module.html#method-i-refine"><code>Module#refine</code></a>
* **Code:**
  ```ruby
  module Tests
    refine Enumerable do # in 2.3: wrong argument type Module (expected Class)
      def tally
        each_with_object(Hash.new(0)) { |el, counter| counter[el] += 1}
      end
    end
  end

  using Tests
  p [1, 3, 1, 2, 1, 3].tally
  # => {1=>3, 3=>2, 2=>1}
  ```

### `Module.used_modules`[](#moduleused_modules)

Returns an array of all modules used in the current scope.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/7418">Feature #7418</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Module.html#method-c-used_modules"><code>Module.used_modules</code></a>
* **Code:**
  ```ruby
  module First
    refine Enumerable do
    end
  end

  module Second
    refine Object do
    end
  end

  module NoRefinements
  end

  p Module.used_modules # => []

  using First
  using Second
  using NoRefinements

  p Module.used_modules # => [First, Second] -- note NoRefinements absence
  ```
  Note that order of modules in array returned is not guaranteed.

## Core[](#core)

### `Warning` module[](#warning-module)

New single-method module was introduced, meant to be overridden in order to control warnings issued by Ruby.

* **Reason:**
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12299">Feature #12299</a>
* **Documentation:** _(introduced in 2.4, but documented in 2.5)_ <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/Warning.html"><code>Warning</code></a>
* **Code:**
  ```ruby
  def Warning.warn(msg)
    puts ".warn called with: #{msg.inspect}"
  end

  X = 1
  X = 2
  # Prints:
  #   .warn called with: "<location>: warning: already initialized constant X\n"
  #   .warn called with: "<location>: warning: previous definition of X was here\n"
  ```
* **Follow-ups:**
  * Surprisingly as at may be, `Kernel#warn` haven't been changed to call `Warning.warn` in 2.4, but it was [fixed in 2.5](2.5.html#warn-call-warningwarn):
    ```ruby
    def Warning.warn(msg)
      puts ".warn called with: #{msg.inspect}"
    end

    warn 'foo', 'bar'
    # Ruby 2.4 prints:
    #  foo
    #  bar
    # Ruby 2.5 prints:
    #  .warn called with: "foo\nbar\n"
    ```
  * In Ruby 2.7, new methods [were added](2.7.html#warning-and-) to `Warning` module allowing control over per-category warning suppression;
  * In Ruby 3.0, category support [was improved](3.0.html#warningwarn-category-keyword-argument).
  * [3.3](3.3.html#new-warning-category-performance): added new warning category: `:performance`.

### `Object#clone(freeze: false)`[](#objectclonefreeze-false)

Allows to receive unfrozen copy of a frozen object.

* **Reason:** Previously, there was no way to receive an unfrozen copy of the frozen object, including its singleton class: `.dup` returns unfrozen object, but doesn't copies the singleton class, while `.clone` copies both (singleton class & frozen state).
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12300">Feature #12300</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Object.html#method-i-clone"><code>Object#clone</code></a>
* **Code:**
  ```ruby
  h = {breed: 'Dog', name: 'Rex'}
  class << h
    def bark
      puts "Bark! Bark!"
    end
  end
  h.freeze
  h.bark    # "Bark! Bark!"

  d = h.dup
  d.frozen? # => false
  d.bark    # NoMethodError: undefined method `bark'

  h2 = h.clone(freeze: false)
  h2[:age] = 8
  h2.bark   # "Bark! Bark!"
  ```
* **Note:** Surprisingly enough, `unfrozen_object.clone(freeze: true)` doesn't make object frozen. See [discussion](https://bugs.ruby-lang.org/issues/16175).
* **Follow-up:** since 3.0:
  * `clone(freeze: true)` [works as expected](3.0.html#objectclonefreeze-true);
  * `freeze:` [argument is passed](3.0.html#objectclone-passes-freeze-argument-to-initialize_clone) to `initialize_clone` so the object could properly freeze/unfreeze its internal data.

### `Comparable#clamp`[](#comparableclamp)

Method to limit any comparable value to `min`—`max` range

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/10594">Feature #10594</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Comparable.html#method-i-clamp"><code>Comparable#clamp</code></a>
* **Code:**
  ```ruby
  123.clamp(0, 20)  # => 20
  -123.clamp(0, 20) # => 0
  18.clamp(0, 20)   # => 18
  ```
* **Follow-up:** In Ruby 2.7, `clamp` also [allows](2.7.html#comparableclamp-with-range) passing range argument, which, especially when combined with [endless](2.6.html#endless-range-1) (2.6) and [beginless](2.7.html#beginless-range) (2.7) ranges, allows to use more powerful and idiomatic code:
  ```ruby
  123.clamp(0..20)  # => 20
  123.clamp(..20)   # => 20
  -123.clamp(0..)   # => 0
  ```

### Numerics[](#numerics)

#### `Fixnum` and `Bignum` are unified into `Integer`[](#fixnum-and-bignum-are-unified-into-integer)

Historically, Ruby had two subclasses of `Integer`: `Fixnum` for numbers that fit into [machine word](https://en.wikipedia.org/wiki/Word_%28computer_architecture%29), and `Bignum` for larger numbers. Since Ruby 2.4, there is only one `Integer`; `Fixnum` and `Bignum` are defined as (deprecated) constants synonymous to it.

* **Reason:** `Fixnum`/`Bignum` separation always been an implementation detail, which led to confusion and sudden bugs, now this detail is hidden by interpreter.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12005">Feature #12005</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Integer.html"><code>Integer</code></a>
* **Code:**
  ```ruby
  # Before 2.4.0
  10.class        # => Fixnum
  (10**100).class # => Bignum

  # 2.4+
  10.class        # => Integer
  (10**100).class # => Integer
  Fixnum
  # warning: constant ::Fixnum is deprecated
  # => Integer
  ```

#### `Numeric#finite?` and `#infinite?`[](#numericfinite-and-infinite)

* **Reason:** The methods were present in `Float` and `BigDecimal`, but not in other numeric classes, which made it harder to write code uniformly processing numbers which may be integer/float/infinite.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12039">Feature #12039</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Numeric.html#method-i-infinite-3F"><code>Numeric#infinite?</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Numeric.html#method-i-finite-3F"><code>Numeric#finite?</code></a>
* **Code:**
  ```ruby
  1.infinite?  # => nil
  1.finite?    # => true
  ```
* **Note:** Notice that `infinite?` returns `nil`/`-1`/`1` (always `nil` for integers), not `true`/`false` as most of other predicate methods. While unusual, it is convenient for checking both for infinity and its sign (+Infinity/-Infinity), and can be treated effectively as `true`/`false` in boolean context.

#### `Integer#digits`[](#integerdigits)

Returns an array of digits of the number.

* **Reason:** Useful for calculating checksums.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12447">Feature #12447</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Integer.html#method-i-digits"><code>Integer.html#digits</code></a>
* **Code:**
  ```ruby
  12345.digits      # => [5, 4, 3, 2, 1] -- digits are returned in lowest-position-first order
  0b11010.digits(2) # => [0, 1, 0, 1, 1] -- optional base can be passed
  ```

#### `ndigits` optional argument for rounding methods[](#ndigits-optional-argument-for-rounding-methods)

Rounding methods of numerics (`ceil`/`floor` etc.) now accept an optional argument to specify how many digits to truncate to. If argument is positive, it means decimal digits, and if it is negative, means tens (the same behavior `#round` had since Ruby 1.9).

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12245">Feature #12245</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Numeric.html#method-i-ceil"><code>Numeric#ceil</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Numeric.html#method-i-floor"><code>Numeric#floor</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Numeric.html#method-i-truncate"><code>Numeric#truncate</code></a> (in fact, `Integer` and `Float` classes are affected, because `Rational` had the same option long ago).
* **Code:**
  ```ruby
  123.4567.ceil(2)  # => 123.46
  123.4567.ceil(0)  # => 124
  123.4567.ceil(-1) # => 130
  ```

#### `half:` option for `#round` method[](#half-option-for-round-method)

For numbers that are exact half, there are several options provided how to round them: up, down, or to the nearest even number. The default behavior kept unchanged (always up).

* **Discussion:** <a class="tracker bug" href="https://bugs.ruby-lang.org/issues/12958">Bug #12958</a> (discussion of rounding behavior), <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12953">Feature #12953</a> (additional rounding options)
* **Documentation:** (feature introduced in 2.4, but comprehensive docs were written in 2.5) <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/Integer.html#method-i-round"><code>Integer#round</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/Float.html#method-i-round"><code>Float#round</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/Rational.html#method-i-round"><code>Rational#round</code></a>
* **Code:**
  ```ruby
  2.5.round # => 3
  2.5.round(half: :down) # => 2
  2.5.round(half: :even) # => 2
  3.5.round(half: :even) # => 4

  25.round(-1, half: :down) # => 20

  (13/2r).round(half: :down) # => 6
  ```

### Strings, symbols and regexps[](#strings-symbols-and-regexps)

> See also [chomp:](#chomp-option-for-string-splitting) option for `String#lines` and `#each_line`, explained in IO section.

#### Unicode case conversions[](#unicode-case-conversions)

All case-conversion methods for `String` and `Symbol` support full Unicode since 2.4.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/10085">Feature #10085</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-downcase"><code>String#downcase</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-upcase"><code>String#upcase</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-capitalize"><code>String#capitalize</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-swapcase"><code>String#swapcase</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-downcase"><code>Symbol#downcase</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-upcase"><code>Symbol#upcase</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-capitalize"><code>Symbol#capitalize</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-swapcase"><code>Symbol#swapcase</code></a>,
* **Code:**
  ```ruby
  'Мамо, ДИВИСЬ, Unicode'.downcase          # => "мамо, дивись, unicode"
  'Мамо, ДИВИСЬ, Unicode'.downcase(:ascii)  # => "Мамо, ДИВИСЬ, unicode"  -- ASCII-only processing (old behavior)

  'Straße'.downcase                         # => "straße"
  'Straße'.downcase(:fold)                  # => "strasse"  -- Unicode case folding

  'TURKIC'.downcase                         # => "turkic"
  'TURKIC'.downcase(:turkic)                # => "turkıc" -- Turkic-specific "dotless i" conversion
  ```

#### `String.new(capacity: size)`[](#stringnewcapacity-size)

When string is created for usage as a mutable buffer for some large textual data, now expected size could be specified, thus optimizing memory allocations.

* **Reason:** Ruby's mutable strings, when used for sequential building of some large text, cause constant reallocations of bigger and bigger memory buffer. By specifying expected capacity beforehand, one can avoid this reallocations.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12024">Feature #12024</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-c-new"><code>String.new</code></a>
* **Code:**
  ```ruby
  s = String.new(capacity: 10_000_000)
  # => "" -- it is still just an empty string, but internal buffer is already allocated large
  ```
* **Note:** Be careful about subtle difference in encoding, when constructing an empty string:
  ```ruby
  # Without source provided, default encoding is ASCII
  String.new(capacity: 10_000_000).encoding # => #<Encoding:ASCII-8BIT>

  # When explicitly constructed from empty string, has this string's encoding (defautl to UTF-8 for
  # string literals)
  String.new('', capacity: 10_000_000).encoding # => #<Encoding:UTF-8>
  ```

#### `#casecmp?`[](#casecmp)

In addition to long-existing case-insensitive comparison method `String#casecmp(other)` (returning `-1`, `0`, `1` like `<=>`), new boolean `String#casecmp?` and `Symbol#casecmp?` were added.

* **Discussion:** [Feature #](https://bugs.ruby-lang.org/issues/)
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-casecmp-3F"><code>String#casecmp?</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-casecmp-3F"><code>Symbol#casecmp?</code></a>
* **Code:**
  ```ruby
  'test'.casecmp?('Test') # => true
  'test'.casecmp?('Tset') # => false
  'test'.casecmp?(:Test)  # TypeError: no implicit conversion of Symbol into String
  ```
* **Follow-up:** In Ruby 2.5, behavior on incompatible types was [changed](https://rubyreferences.github.io/rubychanges/2.5.html#casecmp-and-casecmp-return-nil-for-non-string-arguments) to return `nil`, like `==` does:
  ```ruby
  'test' == :Test # => nil
  'test'.casecmp?(:Test) # => nil
  ```
* **Notes:** It was proposed (in the discussion above), but never implemented to allow passing options to `casecmp?`, making it more precise by specifying locales. Currently, some local characters can produce unexpected results:
  ```ruby
  'ı'.casecmp?('I') # => false, though it is Turkish small dotless "I"
  # ...but...
  'ı'.upcase == 'I' # => true
  ```

#### `String#concat` and `#prepend` accept multiple arguments[](#stringconcat-and-prepend-accept-multiple-arguments)

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12333">Feature #12333</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-concat"><code>String#concat</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-prepend"><code>String#prepend</code></a>
* **Code:**
  ```ruby
  "Hello, ".concat('Judy', ', ', 'John', ' and ', 'Paul')
  # => "Hello, Judy, John and Paul"
  'file.mp3'.prepend('dir1/', 'dir2/')
  # => "dir1/dir2/file.mp3"
  ```

#### `String#unpack1`[](#stringunpack1)

Just a shortcut for `unpack(...).first`

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12752">Feature #12752</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-unpack1"><code>String#unpack1</code></a>
* **Code:**
  ```ruby
  # Previously, you only could've get an Array
  "\x80".unpack('C')  # => [128]

  # Since 2.4
  "\x80".unpack1('C') # => 128
  ```
* **Follow-up:** In Ruby 3.1, [a second argument was added](3.1.html#stringunpack-and-unpack1-offset-argument) to `unpack1` (and `unpack`) allowing to unpack values from the middle of the string.

#### `#match?` method[](#match-method)

New boolean methods for checking if some pattern matches some string/symbol.

* **Reason:** In the (frequent) situation when only "matches or not" is important, boolean `match?` is more readable; also, it is more effective because doesn't set global variables (in case of `Regexp#match?`) and doesn't construct `MatchData`.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/8110">Feature #8110</a>, <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12898">Feature #12898</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Regexp.html#method-i-match-3F"><code>Regexp#match?</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/String.html#method-i-match-3F"><code>String#match?</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Symbol.html#method-i-match-3F"><code>Symbol#match?</code></a>
* **Code:**
  ```ruby
  # before 2.4
  if username =~ /^Admin/

  # Ruby 2.4:
  if username.match?(/^Admin/)

  # Also supports second parameter: position to search matches from:
  if username.match?(/:admin/, 3)
  ```

#### `MatchData`: better support for named captures[](#matchdata-better-support-for-named-captures)

`MatchData#named_captures` returns the hash of `{capture_name => captured string}`; `MatchData#values_at` supports named captures.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11999">Feature #11999</a>, <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/9179">Feature #9179</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/MatchData.html#method-i-named_captures"><code>MatchData#named_captures</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/MatchData.html#method-i-values_at"><code>MatchData#values_at</code></a>
* **Code:**
  ```ruby
  m = 'Serhii Zhadan'.match(/^((?<first>.+?) (?<last>.+?))$/)
  # => #<MatchData "Serhii Zhadan" first:"Serhii" last:"Zhadan">
  m.named_captures
  # => {"first"=>"Serhii", "last"=>"Zhadan"}
  m.values_at(:first, :last) # symbols are supported, too
  # => ["Serhii", "Zhadan"]
  m.values_at(0, :first, 2) # as well as a mix of named and numbered
  # => ["Serhii Zhadan", "Serhii", "Zhadan"]
  ```

### Collections[](#collections)

#### `Enumerable#chunk` without a block returns an `Enumerator`[](#enumerablechunk-without-a-block-returns-an-enumerator)

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/2172">Feature #2172</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Enumerable.html#method-i-chunk"><code>Enumerable#chunk</code></a>
* **Code:**
  ```ruby
  ('a'..'k').chunk
  # => #<Enumerator: "a".."k":chunk>

  # Example of usage:
  ('a'..'k').chunk.with_index { |e, i| (i % 3).zero? }.to_a
  # => [
  #  [true, ["a"]],
  #  [false, ["b", "c"]],
  #  [true, ["d"]],
  #  [false, ["e", "f"]], ...
  ```

#### `#sum`[](#sum)

`Enumerable#sum` was implemented as a core alternative of too common `reduce(:+)`

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12217">Feature #12217</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Enumerable.html#method-i-sum"><code>Enumerable#sum</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Array.html#method-i-sum"><code>Array#sum</code></a>
* **Code:**
  ```ruby
  (1..5).sum # => 15
  (1..5).sum { |x| x ** 2} # => 55

  # Unlike reduce(:+), initial value is implicitly 0, so...
  [].reduce(:+)  # => nil
  [].sum         # => 0

  ('a'..'f').reduce(:+) # => "abcdef"
  ('a'..'f').sum        # TypeError: String can't be coerced into Integer
  ('a'..'f').sum('')    # => "abcdef"
  ```
* **Note:** Separate implementation of `Array#sum` is provided for efficiency. Important thing to note is it doesn't rely on `Array#each` method:
  ```ruby
  class MyAry < Array
    def each(&block)
      super { |val| yield val ** 2 }
    end
  end

  MyAry.new([1, 2, 3, 4, 5]).sum # => 15, not affected by reimplmented #each
  ```

#### `#uniq`[](#uniq)

`#uniq` method, previously present only in `Array`, now available for `Enumerable` and `Enumerator::Lazy`

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11090">Feature #11090</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Enumerable.html#method-i-uniq"><code>Enumerable#iniq</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Enumerator/Lazy.html#method-i-uniq"><code>Enumerator::Lazy#uniq</code></a>
* **Code:**
  ```ruby
  {a: 1, b: 2, c: 1, d: 2, e: 1}.uniq { |k, v| v }
  #  => [[:a, 1], [:b, 2]]

  File.open('very_large_log.log').each_line.lazy.uniq { |ln| ln.scan(/Date: (\S+):/) }.take(10)
  # => first 10 of first-line-of-day
  ```

#### `Array#max` and `#min`[](#arraymax-and-min)

`#max` and `#min` methods of `Enumerable` reimplemented in `Array` for speed.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12172">Feature #12172</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Array.html#method-i-max"><code>Array#max</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Array.html#method-i-max"><code>Array#min</code></a>
* **Note:** Beware that custom reimplementation of `Enumerable#max` and `#min` are now ignored for arrays; and that `Array`'s implementation doesn't use `#each` method.

#### `Array#concat` takes multiple arguments[](#arrayconcat-takes-multiple-arguments)

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12333">Feature #12333</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Array.html#method-i-concat"><code>Array#concat</code></a>
* **Code:**
  ```ruby
  a = [1, 2]
  a.concat([3, 4], [5, 6]) # => [1, 2, 3, 4, 5, 6]
  a # => [1, 2, 3, 4, 5, 6]
  ```

#### `Array#pack(buffer:)`[](#arraypackbuffer)

When provided with optional `buffer:` keyword argument, `Array#pack` uses it as a receiver of data.

* **Reason:** a) pre-allocate the memory for big packed data and b) use the same buffer as a target for several chunks of data
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12754">Feature #12754</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Array.html#method-i-pack"><code>Array#pack</code></a>
* **Code:**
  ```ruby
  # Old way
  [82, 117, 98, 121].pack('C*')        # => "Ruby"
  [32, 105, 115, 32, 99, 111, 111, 108, 33].pack('C*') # => " is cool!"

  # New way
  buffer = String.new(capacity: 30)
  [82, 117, 98, 121].pack('C*', buffer: buffer)
  # => "Ruby"
  buffer # => "Ruby"
  [32, 105, 115, 32, 99, 111, 111, 108, 33].pack('@4C*', buffer: buffer)
  # => "Ruby is cool!"
  buffer # => "Ruby is cool!"

  # Note that if the buffer already has content, the unpacked data is appended to it:
  [32, 73, 115, 32, 105, 116, 63].pack('C*', buffer: buffer)
  # => "Ruby is cool! Is it?"

  # It can be rewritten with explicit offset 0 directive:
  [32, 73, 115, 32, 105, 116, 63].pack('@0C*', buffer: buffer)
  # => " Is it?"
  ```

#### `Hash#compact` and `#compact!`[](#hashcompact-and-compact)

Removes key-value pairs when value is `nil`.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11818">Feature #11818</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Hash.html#method-i-compact"><code>Hash#compact</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Hash.html#method-i-compact-21"><code>Hash#compact!</code></a>
* **Code:**
  ```ruby
  data = {name: 'John', age: 34, occupation: nil}
  data.compact # => {:name=>"John", :age=>34}
  data # => {:name=>"John", :age=>34, :occupation=>nil} -- was not affected
  data.compact! # => {:name=>"John", :age=>34}
  data # => {:name=>"John", :age=>34}
  data.compact! # => nil -- if there were nothing to remove
  ```
* **Note:** Notice the last example: when destructive version haven't changed a hash, it returns `nil` instead of hash itself. It is consistent with behavior of other destructive methods, and allows writing code like this:
  ```ruby
  if data.compact!
    log.info 'Cleaned up the data'
  end
  ```

#### `Hash#transform_values` and `#transform_values!`[](#hashtransform_values-and-transform_values)

Accepts block to transform each value of the hash.

* **Reason:** New method is much easier to write and read, and more effective than
  ```ruby
    hash.map { |key, val| [key, do_something(val)] }.to_h
  ```
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12512">Feature #12512</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Hash.html#method-i-transform_values"><code>Hash#transform_values</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Hash.html#method-i-transform_values-21"><code>Hash#transform_values!</code></a>
* **Code:**
  ```ruby
  h = {x: '10', y: '12', z: '54'}
  h.transform_values(&:to_i) # => {:x=>10, :y=>12, :z=>54}
  h.transform_values!(&:to_i) # => {:x=>10, :y=>12, :z=>54}
  h # => {:x=>10, :y=>12, :z=>54}
  h.transform_values!(&:to_i) # => {:x=>10, :y=>12, :z=>54} -- unlike #compact!, always returns self

  # Without block, returns Enumerator
  h.transform_values
  # => #<Enumerator: {:x=>"10", :y=>"12", :z=>"54"}:transform_values>

  # Can be useful this way:
  h = {manager: 'Jane', reporter: 'John', qa: 'Jane', developer: 'Abraham'}
  h.transform_values.with_index(1) { |v, i| "#{i}: #{v}" }
  # => {:manager=>"1: Jane", :reporter=>"2: John", :qa=>"3: Jane", :developer=>"4: Abraham"}
  ```
* **Follow-up:** Ruby 2.5 also [added](2.5.html#hashtransform_keys-and-transform_keys) `#transform_keys` and `#transform_keys!`

### Filesystem and IO[](#filesystem-and-io)

#### `chomp:` option for string splitting[](#chomp-option-for-string-splitting)

In all contexts where input is split into lines, or received line-by-line, new optional keyword argument `chomp: true` was added to remove (chomp) line-endings.

* **Reason:** Before this change, all line-by-line operations should've included `chomp` as a separate operations when line ending is not needed, which turned out to be most o the cases.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12553">Feature #12553</a>
* **Documentation** (feature introduced in 2.4, but comprehensive docs were written in 2.5-2.6):
  * <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/String.html#method-i-each_line"><code>String#each_line</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.6/String.html#method-i-lines"><code>String#lines</code></a>,
  * <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/IO.html#method-i-gets"><code>IO#gets</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/IO.html#method-i-readline"><code>IO#readline</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/IO.html#method-i-readline"><code>IO#readlines</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/IO.html#method-c-foreach"><code>IO.foreach</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.5.0/IO.html#method-c-readlines"><code>IO.readlines</code></a>,
  * _Standard library: methods of the class affected, but no documentation for the change:_ <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/stringio/rdoc/StringIO.html#method-i-gets"><code>StringIO#gets</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/stringio/rdoc/StringIO.html#method-i-each_line"><code>StringIO#each_line</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/stringio/rdoc/StringIO.html#method-i-readlines"><code>StringIO#readlines</code></a>
* **Code:**
  ```ruby
  # The effect is the same with String, IO and StringIO, so we are demonstrating just one example:
  require 'stringio'
  io = StringIO.new("foo\nbar\nbaz\n")
  io.gets              # => "foo\n"
  io.gets(chomp: true) # => "bar"
  ```
* **Notes:** What is chomped (and what is lines split on) is controlled by `$/` global variable (dubbed `$RS` or `$INPUT_RECORD_SEPARATOR` by `English` module). While quite esoteric by today's standards, it could be really useful for one-off scripts that work with specific data:
  ```ruby
  records = <<~DATA
  First line
  $$$
  Second line
  $$$
  Third line
  DATA

  $/ = "\n$$$\n"
  records.each_line(chomp: true).to_a # => ["First line", "Second line", "Third line"]

  # The same effect, though, can be achieve with normal `separator` method argument:
  records.each_line("\n$$$\n", chomp: true).to_a # => ["First line", "Second line", "Third line"]
  ```

#### `#empty?` method for filesystem objects[](#empty-method-for-filesystem-objects)

New method `empty?` was introduced into several classes to check if the file/directory is empty.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/10121">Feature #10121</a> (`Dir`), <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/9969">Feature #9969</a> (`File`), <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12596">Feature #12596</a> (`Pathname`)
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Dir.html#method-c-empty-3F"><code>Dir#empty?</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/File.html#method-c-empty-3F"><code>File#empty?</code></a>, _(stdlib)_ <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/pathname/rdoc/Pathname.html#method-i-empty-3F"><code>Pathname#empty?</code></a>
* **Code:**
  ```ruby
  Dir.empty?('emptydir')    # => true
  Dir.empty?('nonemptydir') # => false
  Dir.empty?('nonexistent') # Errno::ENOENT (No such file or directory @ rb_dir_s_empty_p - nonexistent)
  Dir.empty?('file')        # => false

  File.empty?('emptyfile')    # => true
  File.empty?('nonemptyfile') # => false
  File.empty?('nonexistent')  # => false -- unlike Dir.empty?
  File.empty?('dir')          # => false

  require 'pathname'
  Pathname('emptydir').empty?    # => true
  Pathname('emptyfile').empty?   # => true
  Pathname('nonexistent').empty? # => false
  Pathname('nonempty').empty?    # => false
  ```

### `Thread#report_on_exception` and `Thread.report_on_exception`[](#threadreport_on_exception-and-threadreport_on_exception)

Global and thread-local boolean flag to set what should the thread do when ended with exception: die silently (default, old behavior) or print the exception and backtrace to `$stderr`.

* **Reason:** Threads silently dying without any indication could be a lot of confusion, and before this feature top-level exception reporting for each thread should've been implemented manually.
* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/6647">Feature #6647</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Thread.html#method-c-report_on_exception"><code>Thread.report_on_exception</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Thread.html#method-c-report_on_exception-3D"><code>Thread.report_on_exception=</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Thread.html#method-i-report_on_exception"><code>Thread#report_on_exception</code></a>, <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Thread.html#method-i-report_on_exception-3D"><code>Thread#report_on_exception</code></a>
* **Code:**
  ```ruby
  Thread.new { puts 1 / 0 }
  sleep(1)
  # => nothing happens, thread is dead
  Thread.report_on_exception = true
  Thread.new { puts 1 / 0 }
  sleep(1)
  # #<Thread:0x0055b070475fb0> terminated with exception:
  # in `/': divided by 0 (ZeroDivisionError)

  # Or instance-level method:
  t = Thread.new { sleep(1); puts 1 / 0 }
  t.report_on_exception = false # silence it again
  sleep(1)
  # => Thread dies in a sad silence.
  ```
* **Follow-up:** Since 2.5, `Thread.report_on_exception` is `true` by default.

### `TracePoint#callee_id`[](#tracepointcallee_id)

Returns an actual name of the method being called, even if aliased.

* **Discussion:** <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12747">Feature #12747</a>
* **Documentation:** <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/TracePoint.html#method-i-callee_id"><code>TracePoint#callee_id</code></a>
* **Code:**
  ```ruby
  tp = TracePoint.new(:call) { |point| p [point.method_id, point.callee_id]}

  def real
  end

  alias aliased real

  tp.enable { aliased } # prints method id and callee id: [:real, :aliased]
  ```

## Stdlib[](#stdlib)

* `Set`: <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/set/rdoc/Set.html#method-i-compare_by_identity"><code>#compare_by_identity</code></a> and <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/set/rdoc/Set.html#method-i-compare_by_identity-3F"><code>#compare_by_identity?</code></a> methods added, behaving the same way as (existing since 1.9) <a class="ruby-doc" href="https://ruby-doc.org/core-2.4.0/Hash.html#method-i-compare_by_identity"><code>Hash#compare_by_identity</code></a>: only elements being the same object (same `#object_id`) are considered same set element. Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12210">Feature #12210</a>
* <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/csv/rdoc/CSV.html#method-c-new"><code>CSV.new</code></a>: Add a `liberal_parsing` option, allowing to (try to) parse not-completely-valid CSV. Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11839">Feature #11839</a>
* `Binding#irb` start a REPL session like `binding.pry`.Follow-up: since [Ruby 2.5](https://rubyreferences.github.io/rubychanges/2.5.html#ruby-development-and-introspection), `require 'irb'` is not necessary for the feature to work, it is done automatically.
* <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/logger/rdoc/Logger.html#method-c-new"><code>Logger.new</code></a> adds keyword arguments `level:`, `progname:`, `datetime_format:`, `formatter:`, `shift_period_suffix:`. The latter allows specifying suffix for filenames on log rotation. Discussions: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12224">Feature #12224</a> (keyword args), <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/10772">Feature #10772</a> (`shift_period_suffix`).
* <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/net/http/rdoc/Net/HTTP.html#method-c-post"><code>Net::HTTP.post</code></a> shortcut method. Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12375">Feature #12375</a>
* `Net::FTP`:
  * Support TLS.
  * Support hash style options for <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/net/ftp/rdoc/Net/FTP.html#method-c-new"><code>Net::FTP.new</code></a>. While not reflected in the docs, "old" way (as it <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.3.0/libdoc/net/ftp/rdoc/Net/FTP.html#method-c-new"><code>was</code></a> before 2.4, with separate args for user, password etc.) still works for backwards compatibility.
  * <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/net/ftp/rdoc/Net/FTP.html#method-i-status"><code>Net::FTP#status</code></a>: optional argument `pathname` (`STAT path` "is analogous to the "list" command, except that data shall be transferred over the control connection"). Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12965">Feature #12965</a>
* `OptionParser` (optparse): <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/optparse/rdoc/OptionParser.html#method-i-parse-21"><code>OptionParser#parse!</code></a> and similar methods add `into:` option to parse, greatly simplifying trivial case of "parse into hash". Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/11191">Feature #11191</a>
  ```ruby
  require 'optparse'
  opts = OptionParser.new do |o|
    o.on '-p', '--port=PORT', 'port', Integer
    o.on '-v', '--verbose'
  end

  result = {}
  opts.parse!(%w[-p 8080 -v], into: result)
  p result # => {:port=>8080, :verbose=>true}
  ```

* `Readline`: <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/readline/rdoc/Readline.html#method-c-quoting_detection_proc"><code>::quoting_detection_proc</code></a> and <a class="ruby-doc" href="https://ruby-doc.org/stdlib-2.4.0/libdoc/readline/rdoc/Readline.html#method-c-quoting_detection_proc-3D"><code>quoting_detection_proc=</code></a> to specify callable object (`Proc` or anything responding to `#call`), customizing the decision "if in this line, character in that position is quoted or not". It is standard functionality of GNU readline which was not previously exposed by Ruby wrapper. Discussion: <a class="tracker feature" href="https://bugs.ruby-lang.org/issues/12659">Feature #12659</a>

### Libraries promoted to default/bundled gems[](#libraries-promoted-to-defaultbundled-gems)

> **[stdgems.org](https://stdgems.org/)** project has a nice explanations of default and bundled gems concepts, as well as a list of currently gemified libraries.

"For the rest of us":
* _default_ means libraries development extracted into separate GitHub repositories, and they are just packaged with main Ruby before release. It means you can do issue/PR to any of them independently, without going through more tough development process of the core Ruby;
* _bundled_ means libraries development also extracted, and they are _not_ packaged with Ruby distribution, just automatically installed with it.

Libraries that became default in 2.4:

* <a class="github" href="https://github.com/ruby/openssl">openssl</a>
* <a class="github" href="https://github.com/ruby/webrick">webrick</a>

Libraries that became bundled in 2.4:

* <a class="github" href="https://github.com/ruby/xmlrpc">xmlrpc</a>

**Follow-up:**
* [16 more libraries](2.5.html#libraries-promoted-to-default-gems) gemified in 2.5;
* [14 more libraries](2.6.html#libraries-promoted-to-default-gems) gemified in 2.6;
* [16 more libraries](2.7.html#libraries-promoted-to-default-gems) gemified in 2.7, and [6 just dropped](2.7.html#libraries-excluded-from-the-standard-library) from the standard library;
* [34 (!) more libraries](3.0.html#libraries-promoted-to-default-gems) gemified in 3.0, and [3 more just dropped](3.0.html#libraries-promoted-to-default-gems) from the standard library (including xmlrpc and webrick).
